#!/usr/bin/perl
#
# $Id$
#
# (C) IBCL Informatik-Buero e.U. 2010
# (C) ITEG IT-Engineers GmbH 2016
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#   http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

use strict;

use Getopt::Std;
use Net::SMTP;
use Email::MIME;
use IO::All;
use Expect;
use Time::HiRes qw( gettimeofday tv_interval );
use Dpkg::Version;


sub initPBAFunctions() { 

  # assertions:
  # $main::role set to "archiver" or "builder"
  # $main::optionList set to "c:dqv" or more;


  # detect hostname
  $main::hostname = $ENV{"HOSTNAME"};
  if ( $main::hostname eq "" ) {
    $main::hostname = `hostname -f`;
  }
  chomp $main::hostname;

  $main::version = "1.6.34-3";
  if (length($main::scriptName) > 0) {
    $main::versionMessage = $main::scriptName." (pba-".$main::role." ".$main::version.")\n";
  } else {
    $main::versionMessage = "pba-".$main::role." ".$main::version."\n";
  }

  $main::proxy = $ENV{"HTTP_PROXY"};

  %main::opts                         = ();
  $Getopt::Std::STANDARD_HELP_VERSION = 1;
  getopts( $main::optionList, \%main::opts );

  #$main::user = $ENV{"USER"};
  # real user $< over effective user $>, because sudo pba-builder ... is only typical sudo usage:
  my @passwdParts = getpwuid($<);
  $main::userHome = $passwdParts[7];

  $main::configDir = $main::opts{"c"};
  if ( $main::configDir eq "" ) {
    $main::configDir = "/etc/pba";
  }
  $main::configFile = $main::configDir."/".$main::role.".conf";
  $main::userConfigFile = $main::userHome."/.pba/".$main::role.".conf";

  $main::verbosity = 1;
  if ( $main::opts{"d"} ) {
    $main::verbosity += 2;
  }
  if ( $main::opts{"q"} ) {
    $main::verbosity = 0;
  }
  if ( $main::opts{"v"} ) {
    $main::verbosity += 1;
  }
  if ($main::verbosity > 0) {
    main::VERSION_MESSAGE();
  }

  # set default params
  %main::configParams = ();

  @main::filesToCleanup = ();
  @main::dirsToCleanup = ();
  @main::mountsToUnmount = ();
}


sub parseConfig() {

  my $formerPwd = pwd();

  if ( -f $main::userHome."/.gnupg/gpg.conf" ) {
    ($main::verbosity > 1) && print "\nParsing ".$main::userHome."/.gnupg/gpg.conf for default-key over GPGKeyId\n\n"; 
    open( CONFFILE, "<".$main::userHome."/.gnupg/gpg.conf" ) || cleanupAndDie("FATAL: Cannot read config file ".$main::userHome."/.gnupg/gpg.conf\n");
    while ( my $curLine = <CONFFILE> ) {
      chomp $curLine;
      next if ( $curLine =~ m/^\s*[#;]/ );
      next if ( $curLine =~ m/^\s*$/ );
      my ( $pn, $pv ) = split / +/, $curLine, 2;
      if ($pn eq "default-key") {
        $main::configParams{"GPGKeyId"} = $pv;
        ($main::verbosity > 1) && print "\nUsing default-key from ".$main::userHome."/.gnupg/gpg.conf as default GPGKeyId: '".$pv."\n\n"; #' (line '".$curLine."')\n\n"; 
      }
    }
    close CONFFILE;
  } else {
    ($main::verbosity > 2) && print "\nNot parsing non-existent optional config file ".$main::userHome."/.gnupg/gpg.conf\n\n"; 
  }

  # read config file
  ( $main::configFile eq "") && cleanupAndDie("FATAL: Empty configFile specification.\n");
  if ( $main::configFile ne "" ) {
    ($main::verbosity > 1) && print "\nParsing config file $main::configFile\n\n"; 
    open( CONFFILE, "<$main::configFile" ) || cleanupAndDie("FATAL: Cannot read config file ".$main::configFile."\n");
    while ( my $curLine = <CONFFILE> ) {
      chomp $curLine;
      next if ( $curLine =~ m/^\s*[#;]/ );
      next if ( $curLine =~ m/^\s*$/ );
      my ( $pn, $pv ) = split /: +/, $curLine, 2;
      $main::configParams{$pn} = $pv;
    }
    close CONFFILE;
  }
  
  if ( $main::userConfigFile ne "" ) {
    if ( -f $main::userConfigFile ) {
      ($main::verbosity > 1) && print "\nParsing config file $main::userConfigFile\n\n"; 
      open( CONFFILE, "<$main::userConfigFile" ) || cleanupAndDie("FATAL: Cannot read config file ".$main::userConfigFile."\n");
      while ( my $curLine = <CONFFILE> ) {
        chomp $curLine;
        next if ( $curLine =~ m/^\s*[#;]/ );
        next if ( $curLine =~ m/^\s*$/ );
        my ( $pn, $pv ) = split /: +/, $curLine, 2;
        $main::configParams{$pn} = $pv;
      }
      close CONFFILE;
    } else {
      ($main::verbosity > 2) && print "\nNot parsing non-existent optional config file $main::userConfigFile\n\n"; 
    }
  }
 
  # print config if debugging
  if ( $main::verbosity > 3 ) {
    print "Actual $main::role configuration:\n";
    foreach my $paramName(sort(keys(%main::configParams))) {
      print "$paramName = '".$main::configParams{$paramName}."'\n";  
    }
    print "\n"; 
  }

  # check of config moved to programs
  
  # read repositories.d/*
  @main::repositories = ();
  %main::repositoryAutoFeed = ();
  %main::repositoryBaseURIs = ();
  %main::repositoryDistributions = ();
  %main::repositoryFeedHostDirs = ();
  %main::repositoryLongNames = ();
  %main::repositoryRoots = ();
  %main::repositoryTypes = ();
  %main::repositoryGPGKeyName = ();
  my @repositoryFiles = ();
  if (-d $main::configDir."/repositories.d/") {
    chdir $main::configDir."/repositories.d/" || cleanupAndDie("FATAL: Cannot chdir to ".$main::configDir."/repositories.d/\n");
    foreach my $file (glob("*")) {
      push @repositoryFiles, $main::configDir."/repositories.d/".$file;
    }
  }
  if (-d $main::userHome."/.pba/repositories.d/") {
    chdir $main::userHome."/.pba/repositories.d/" || cleanupAndDie("FATAL: Cannot chdir to ".$main::userHome."/.pba/repositories.d/\n");
    foreach my $file (glob("*")) {
      push @repositoryFiles, $main::userHome."/.pba/repositories.d/".$file;
    }
  }
  if (scalar(@repositoryFiles) != 0) {
    foreach my $repositoryFile (@repositoryFiles) {
      next if (! -f $repositoryFile);
      if ($repositoryFile =~ /(\.dpkg|\.template|.disabled|~$)/i) {
        ($main::verbosity > 1) && print "Ignoring repository file $repositoryFile\n";
        next;
      }
      my $repository = $repositoryFile;
      $repository =~ s/.*\///g;
      push @main::repositories, $repository;
      $main::repositoryAutoFeed{$repository} = "";
      $main::repositoryBaseURIs{$repository} = "";
      $main::repositoryDistributions{$repository} = "";
      $main::repositoryFeedHostDirs{$repository} = "";
      $main::repositoryLongNames{$repository} = $repository;
      $main::repositoryRoots{$repository} = "";
      $main::repositoryTypes{$repository} = "";
      $main::repositoryGPGKeyName{$repository} = "";
      open( INFILE, "<$repositoryFile" ) || cleanupAndDie("FATAL: Cannot read ".$repositoryFile."\n");
      ($main::verbosity > 1) && print "Parsing Distribution file ".$repositoryFile."\n";
      my $lineNr = 0;
      while ( my $curLine = <INFILE> ) {
        chomp $curLine;
        $lineNr++;
        next if ( $curLine =~ m/^\s*[#;]/ );
        next if ( $curLine =~ m/^\s*$/ );
        my ( $pn, $pv ) = split /: +/, $curLine, 2;
        if ( $pn eq "AutoFeed" ) {
          $main::repositoryAutoFeed{$repository} = $pv;
          next;
        }
        if ( $pn eq "BaseURI" ) {
          $main::repositoryBaseURIs{$repository} = $pv;
          next;
        }
        if ( $pn eq "Distributions" ) {
          $main::repositoryDistributions{$repository} = $pv;
          next;
        }
        if ( $pn eq "FeedHostDir" ) {
          $main::repositoryFeedHostDirs{$repository} = $pv;
          next;
        }
        if ( $pn eq "LongName" ) {
          $main::repositoryLongNames{$repository} = $pv; 
          next;
        }
        if ( $pn eq "RepositoryRoot" ) {
          $main::repositoryRoots{$repository} = $pv;
          next;
        }
        if ( $pn eq "RepositoryType" ) {
          $main::repositoryTypes{$repository} = $pv;
          next;
        }
        if ( $pn eq "GPGKey" ) {
          $main::repositoryGPGKeyName{$repository} = $pv;
          next;
        }
        ($main::verbosity > 1) && print "WARNING: Unkown parameter '".$pn."' in ".$repositoryFile.":".$lineNr.".\n";
      }
      close INFILE;
      # these could be move to the programs but seem generic enough
      if ($main::repositoryAutoFeed{$repository} eq "") {
        if ($main::repositoryFeedHostDirs{$repository} eq "") {
          $main::repositoryAutoFeed{$repository} = unifyBool($main::repositoryAutoFeed{$repository}, "false");
        } else {
          $main::repositoryAutoFeed{$repository} = unifyBool($main::repositoryAutoFeed{$repository}, "true");
        }
      }
    }
  } else {
    if ( ($main::scriptName eq "pba-builder") or ($main::scriptName eq "pba-archiver") ) {
      print "ERROR: No distributions.d directory ".$main::configDir."/distributions.d/"."\n";
    }
  }

  # list repositories
  if ($main::verbosity > 3) {
    print "\nRepositories:\n";
    foreach my $repository (@main::repositories) {
      print "  ".$repository.":\n";
      print "    AutoFeed: ".$main::repositoryAutoFeed{$repository}."\n";  
      print "    BaseURI: ".$main::repositoryBaseURIs{$repository}."\n";
      print "    Distributions: ".$main::repositoryDistributions{$repository}."\n";  
      print "    Feed host:dir: ".$main::repositoryFeedHostDirs{$repository}."\n";  
      print "    Long name: ".$main::repositoryLongNames{$repository}."\n";  
      print "    Repository root: ".$main::repositoryRoots{$repository}."\n";  
    }
    print "\n"; 
  }
  
  # read distribution.d/*
  @main::distributions = ();
  %main::distributionLongNames = ();
  %main::distributionFamilies = ();
  %main::distributionMirrors = ();
  %main::distributionArchitectures = ();
  my @distributionFiles = ();
  if (-d $main::configDir."/distributions.d/") {
    chdir $main::configDir."/distributions.d/" || cleanupAndDie("FATAL: Cannot chdir to ".$main::configDir."/distributions.d/\n");
    foreach my $file (glob("*")) {
      push @distributionFiles, $main::configDir."/distributions.d/".$file;
    }
  }
  if (-d $main::userHome."/.pba/distributions.d/") {
    chdir $main::userHome."/.pba/distributions.d/" || cleanupAndDie("FATAL: Cannot chdir to ".$main::userHome."/.pba/distributions.d/\n");
    foreach my $file (glob("*")) {
      push @distributionFiles, $main::userHome."/.pba/distributions.d/".$file;
    }
  }
  if (scalar(@distributionFiles) != 0) {
    foreach my $distributionFile (@distributionFiles) {
      next if (! -f $distributionFile);
      if ($distributionFile =~ /(\.dpkg|\.template|.disabled|~$|-)/i) {
        ($main::verbosity > 1) && print "Ignoring distribution file $distributionFile\n";
        next;
      }
      my $distribution = $distributionFile;
      $distribution =~ s/.*\///g;
      push @main::distributions, $distribution;
      $main::distributionArchitectures{$distribution} = "amd64"; # i386
      $main::distributionFamilies{$distribution} = "";
      $main::distributionLongNames{$distribution} = $distribution;
      $main::distributionMirrors{$distribution} = "";
      $main::distributionBootstrapHints{$distribution} = "";
      open( INFILE, "<$distributionFile" ) || cleanupAndDie("FATAL: Cannot read ".$distributionFile."\n");
      ($main::verbosity > 1) && print "Parsing Distribution file ".$distributionFile."\n";
      my $lineNr = 0;
      while ( my $curLine = <INFILE> ) {
        chomp $curLine;
        $lineNr++;
        next if ( $curLine =~ m/^\s*[#;]/ );
        next if ( $curLine =~ m/^\s*$/ );
        my ( $pn, $pv ) = split /: +/, $curLine, 2;
        if ( $pn eq "Architectures" ) {
          $main::distributionArchitectures{$distribution} = $pv;
          next;
        }
        if ( $pn eq "BootstrapHint" ) {
          $main::distributionBootstrapHints{$distribution} = $pv;
          next;
        }
        if ( $pn eq "Family" ) {
          $main::distributionFamilies{$distribution} = $pv;
          next;
        }
        if ( $pn eq "Mirror" ) {
          $main::distributionMirrors{$distribution} = $pv;
          next;
        }
        if ( $pn eq "LongName" ) {
          $main::distributionLongNames{$distribution} = $pv; 
          next;
        }
        ($main::verbosity > 1) && print "WARNING: Unkown parameter '".$pn."' in ".$distributionFile.":".$lineNr.".\n";
      }
      close INFILE;
      $main::distributionFamilies{$distribution} = simplifyList($main::distributionFamilies{$distribution});
      ($main::distributionFamilies{$distribution} eq "") && cleanupAndDie("FATAL: No Family specified in ".$distributionFile."\n");
      $main::distributionMirrors{$distribution} = simplifyList($main::distributionMirrors{$distribution});
      ($main::distributionMirrors{$distribution} eq "") && cleanupAndDie("FATAL: No Mirror specified in ".$distributionFile."\n");
    }
  } else {
    if ( ($main::scriptName eq "pba-builder") or ($main::scriptName eq "pba-archiver") ) {
      print "ERROR: No distributions.d directory ".$main::configDir."/distributions.d/"."\n";
    }
  }

  # list distributions
  if ($main::verbosity > 3) {
    print "\nDistributions:\n";
    foreach my $distribution (@main::distributions) {
      print "  ".$distribution.":\n";
      print "    Long name: ".$main::distributionLongNames{$distribution}."\n";  
      print "    Family: ".$main::distributionFamilies{$distribution}."\n";  
      print "    Mirror: ".$main::distributionMirrors{$distribution}."\n";  
      print "    Architectures: ".$main::distributionArchitectures{$distribution}."\n";  
    }
    print "\n"; 
  }

  # read buildhosts.d/*
  @main::buildhosts = ();
  %main::buildhostFeedHostDirs = ();
  %main::buildhostLongNames = ();
  my @buildhostFiles = ();
  if (-d $main::configDir."/buildhosts.d/") {
    chdir $main::configDir."/buildhosts.d/" || cleanupAndDie("FATAL: Cannot chdir to ".$main::configDir."/buildhosts.d/\n");
    foreach my $file (glob("*")) {
      push @buildhostFiles, $main::configDir."/buildhosts.d/".$file;
    }
  }
  if (-d $main::userHome."/.pba/buildhosts.d/") {
    chdir $main::userHome."/.pba/buildhosts.d/" || cleanupAndDie("FATAL: Cannot chdir to ".$main::userHome."/.pba/buildhosts.d/\n");
    foreach my $file (glob("*")) {
      push @buildhostFiles, $main::userHome."/.pba/buildhosts.d/".$file;
    }
  }
  foreach my $buildhostFile (@buildhostFiles) {
    next if (! -f $buildhostFile);
    if ($buildhostFile =~ /(\.dpkg|\.template|.disabled|~$)/i) {
      ($main::verbosity > 1) && print "Ignoring buildhost file $buildhostFile\n";
      next;
    }
    my $buildhost = $buildhostFile;
    $buildhost =~ s/.*\///g;
    push @main::buildhosts, $buildhost;
    $main::buildhostFeedHostDirs{$buildhost} = "";
    $main::buildhostLongNames{$buildhost} = $buildhost;
    open( INFILE, "<$buildhostFile" ) || cleanupAndDie("FATAL: Cannot read ".$buildhostFile."\n");
    ($main::verbosity > 1) && print "Parsing Distribution file ".$buildhostFile."\n";
    my $lineNr = 0;
    while ( my $curLine = <INFILE> ) {
      chomp $curLine;
      $lineNr++;
      next if ( $curLine =~ m/^\s*[#;]/ );
      next if ( $curLine =~ m/^\s*$/ );
      my ( $pn, $pv ) = split /: +/, $curLine, 2;
      if ( $pn eq "FeedHostDir" ) {
        $main::buildhostFeedHostDirs{$buildhost} = $pv;
        next;
      }
      if ( $pn eq "LongName" ) {
        $main::buildhostLongNames{$buildhost} = $pv; 
        next;
      }
      ($main::verbosity > 1) && print "WARNING: Unkown parameter '".$pn."' in ".$buildhostFile.":".$lineNr.".\n";
    }
    close INFILE;
  }

  # list buildhosts
  if ($main::verbosity > 2) {
    print "\nBuild hosts:\n";
    foreach my $buildhost (@main::buildhosts) {
      print "  ".$buildhost.":\n";
      print "    Feed host:dir: ".$main::buildhostFeedHostDirs{$buildhost}."\n";  
      print "    Long name: ".$main::buildhostLongNames{$buildhost}."\n";  
    }
    print "\n"; 
  }
  
  chdir $formerPwd || cleanupAndDie("FATAL: Cannot chdir back to ".$formerPwd."\n");

} # end init


# functions


# for GetOpt:

sub main::HELP_MESSAGE() {
  print $main::helpMessage;
}

sub main::VERSION_MESSAGE() {
  print $main::versionMessage;
}


# for unifying config values


sub unifyBool ( $ $ ) { 
  # $1 = value to parse
  # $2 = default value, i.e. "true", "false"
  # returns "true" or "false"

  my $b = shift;
  my $default = shift;
  ( ($default ne "true") and ($default ne "false") ) && cleanupAndDie("FATAL: unifyBool(\$,\$) called with illegal default value '".$default."'\n");

  $b =~ s/\s//g; 
  if ($b =~ /(true|yes|1)/i) {
    return "true";
  }
  if ($b =~ /(false|no|0)/i) {
    return "false";
  }
  return $default;
}


sub simplifyList ( $ ) {
  # $1 = potentially unclean list string ($, not @)
  # returns list string trimmed and with separations unified to " ", 
  # i.e. ready for:
  # my @myList = split / /, simplyfyList( $myDirtyListString );

  my $s = shift;
  # empty $s is ok!

  $s =~ s/,?\s+/ /g; 
  $s =~ s/^\s+//; 
  $s =~ s/\s+$//; 
  return $s;
}



# directory helpers


sub checkDirectory ( $ ; $ ) {
  # $1 = directory to check and auto-create
  # $2 = mode; optional, defaults to "0755"
  my $directory = shift;
  ($directory eq "") && cleanupAndDie("FATAL: checkDirectory(\$[,\$]) called with empty directory spec.\n");
  my $mode = shift;
  if ($mode eq "") { $mode = "0755"; }

  if ($directory =~ /\s/) {
    cleanupAndDie("FATAL: checkDirectory(): illegal characters in directory: '".$directory."'\n");
  }
  if (-d $directory) { 
    return;
  }
  ($main::verbosity > 2) && print "Auto-creating directory ".$directory.".\n";
  my $command = "/bin/mkdir -m ".$mode." -p ".$directory;
  my $rc = system($command);
  ($rc eq 0) || cleanupAndDie("FATAL: Unable to auto-create directory ".$directory."\nFull command was:\n".$command."\nCurrent directory is: ".pwd()."\n");
}


sub pwd() {
  # returns current directory, chomped
  my $currentDir = `pwd`;
  chomp $currentDir;
  return $currentDir;
}



# file helpers


sub printToFile( $ $ ; $ ) {
  # $1 = filename
  # $2 = text
  # $3 = mode, optional, defaults to 0755
  my $filename = shift;
  my $text = shift;
  my $mode = shift;
  if ($mode eq "") { $mode = 0755; }
  open (TMPFILE, ">".$filename) || cleanupAndDie("FATAL: Cannot write ".$filename."\n");
  print TMPFILE $text;
  close TMPFILE;
  if ($filename =~ /\.(sh|pl)$/) {
    chmod $mode, $filename || cleanupAndDie("FATAL: Cannot chmod ".$filename."\n");
  }
}


sub readCompleteFile( $ ) {
  # $1 = name of text file
  # returns content of file
  my $filename=shift;
  my $text;
  my $defaultLF=$/;
  undef $/;
  open (TMPFILE, "<".$filename) || cleanupAndDie("FATAL: Cannot read ".$filename);
  $text=<TMPFILE>;
  close TMPFILE;
  $/=$defaultLF;
  return $text;
}



# Signature stuff

# for Debian repositories' root Release file
sub signAndChmod ( $ ; $ ) {
  # $1 = file to sign and chmod
  # $2 = mode to chmod signed file and signature to; optional, defaults to 0755
  my $file = shift;
  ($file eq "") && cleanupAndDie("FATAL: signAndChmod(\$[,\$]) called with empty file spec.\n");
  (-f $file) || cleanupAndDie("FATAL: signAndChmod(".$file.",\$]): no such file.\n");
  my $mode = shift;
  if ($mode eq "") { $mode = 0755; }

  # Release.gpg
  my $gpgFile = $file.".gpg";
  (-f $gpgFile) && unlink $gpgFile;
  my $gpgCmd = "gpg";
  if (length($main::gpgHome) > 0) {
    $gpgCmd .= " --homedir ".$main::gpgHome;
  }
  if (length($main::gpgKeyId) > 0) {
    $gpgCmd .= " --default-key ".$main::gpgKeyId;
  }
  $gpgCmd .= " -sb -a -o ".$gpgFile." ".$file;;
  executeOrDie($gpgCmd);

  # InRelease
  my $inFile = $file;
  $inFile =~ s/Release/InRelease/;
  (-f $inFile) && unlink $inFile;
  $gpgCmd = "gpg";
  if (length($main::gpgHome) > 0) {
    $gpgCmd .= " --homedir ".$main::gpgHome;
  }
  if (length($main::gpgKeyId) > 0) {
    $gpgCmd .= " --default-key ".$main::gpgKeyId;
  }
  $gpgCmd .= " --clearsig -a -o ".$inFile." ".$file;;
  executeOrDie($gpgCmd);
  
  chmod $mode, $file, $file.".gpg", $inFile;
}


# check rpm signature with rpm -K, demanding RSA.*OK$ in the output
# $1 = rpm file
sub checkRpmSignature( $ ) {
  my $rpmFile = shift;
  ($rpmFile eq "") && cleanupAndDie("FATAL: checkRpmSignature(\$) called with empty file spec.\n");

  #print "Listing %ENV ... before checking rpm RSA signature ...\n";
  #for (sort keys %ENV) {
  #  print "$_ => $ENV{$_}\n";
  #}
  #print "Listed %ENV ... before checking rpm RSA signature.\n";
  # sometimes HOME is set to /, in which case rpm searches keys in wrong place
  # for now we fix that in surrounding scripts

  my $rpmCmd = "rpm -v";
  if ($main::verbosity > 2) {
    $rpmCmd .= "v";
  }
  $rpmCmd .= " -K ".$rpmFile;
  if ($main::verbosity > 1) { 
    print "Checking rpm RSA signature: ".$rpmCmd."\n";
  }
  my $rpmOutput = `$rpmCmd`;
  if ($rpmOutput =~ m/RSA.*: OK$/m) {
    ($main::verbosity > 1) && print "RSA signature approved of rpm ".$rpmFile."\n";
    return 1;
  }
  ($main::verbosity > 0) && print "RSA signature of rpm ".$rpmFile." FAILED to approve:\n".$rpmOutput."\n\n";
  return 0;
}


# ATTENTION: Obsolete version for rpm up to Debian Jessie
# correctly sign rpm file, even unattended if correctly configured
# $1 = rpm file
sub signRpmJessie( $ ) {
  my $rpmFile = shift;
  ($rpmFile eq "") && cleanupAndDie("FATAL: signRpmJessie(\$) called with empty file spec.\n");

  #executeOrDie("/usr/bin/pba-signrpm.expect ../".$srpmName."\n");

  my $gpgPassphrase = "";
  if (defined($ENV{"GPG_PASSPHRASE"})) {
    $gpgPassphrase = $ENV{"GPG_PASSPHRASE"};
  }

  my $gpgAgentInfo = "";
  if (defined($ENV{"GPG_AGENT_INFO"})) {
    $gpgAgentInfo = $ENV{"GPG_AGENT_INFO"};
  }
  my $gpgUser = "";
  if (defined($ENV{"GPG_USER"})) {
    $gpgUser = $ENV{"GPG_USER"};
  }

  my $rpmCmd = "rpm -addsign";

  # gpg_sign_cmd
  $rpmCmd .= ' -D "%__gpg_sign_cmd %{__gpg} --no-verbose --no-armor --passphrase-fd 3';
  if (length($gpgAgentInfo) > 0) {
    $rpmCmd .= ' --use-agent --gpg-agent-info '.$gpgAgentInfo;
  } else {
    $rpmCmd .= ' --no-use-agent';
  }
  if (length($main::gpgHome) > 0) {
    $rpmCmd .= " --homedir ".$main::gpgHome;
  }
  if (length($main::gpgKeyId) > 0) {
    $rpmCmd .= " --default-key ".$main::gpgKeyId;
  }
  if (length($gpgUser) > 0) {
    $rpmCmd .= ' -u \"%{_gpg_name}\"';
  }
  $rpmCmd .= ' -sbo %{__signature_filename} %{__plaintext_filename}"';

  # gpg_name
  if (length($gpgUser) > 0) {
    $rpmCmd .= ' -D "%_gpg_name '.$gpgUser.'"';
  } else {
    $rpmCmd .= ' -D "%_gpg_name foobar"';
  }

  # gpg_check_password_cmd
  if (length($gpgAgentInfo) > 0) {
    $rpmCmd .= ' -D "%__gpg_check_password_cmd /bin/true"';
  } else {
    $rpmCmd .= ' -D "%__gpg_check_password_cmd %{__gpg} --batch --no-verbose --passphrase-fd 3';
    if (length($main::gpgHome) > 0) {
      $rpmCmd .= " --homedir ".$main::gpgHome;
    }
    if (length($main::gpgKeyId) > 0) {
      $rpmCmd .= " --default-key ".$main::gpgKeyId;
    }
    if (length($gpgUser) > 0) {
      $rpmCmd .= ' -u \"%{_gpg_name}\"';
    }
    $rpmCmd .= ' -so -"';
  }
  $rpmCmd .= " ".$rpmFile;

  ($main::verbosity > 1) && print "signRpmJessie(): Going to execute the following command:\n".$rpmCmd."\n\n";

  my $t0 = [gettimeofday];

  #print "DEBUG after ".tv_interval($t0)."s: Creating \$exp ...\n";
  my $exp = new Expect() || cleanupAndDie("FATAL: Cannot create tty (new Expect() failed).\n");
  #print "DEBUG after ".tv_interval($t0)."s: Created \$exp, spawning ...\n";
  $exp->spawn($rpmCmd) || cleanupAndDie("FATAL: Cannot spawn rpmsign command.\n");
  # on  $exp->
  #print "DEBUG after ".tv_interval($t0)."s: Spawned, expecting ...\n";
  my $nrPhrases = 0;
  my $gotEof = 0;
  my $gotTimeout = 0;
  $exp->expect(10,
    [ "Enter pass phrase", 
      sub { 
        $nrPhrases += 1; 
        my $subExp = shift; 
        $subExp->send($gpgPassphrase."\r"); 
        exp_continue; 
      } 
    ],
    [ eof => 
      sub { 
        $gotEof = 1;
      } 
    ],
    [ timeout => 
      sub { 
        $gotTimeout = 1; 
      }
    ]
  );
  #print "DEBUG after ".tv_interval($t0)."s: Expected, retrieving lastError.\n";
  my $lastError = $exp->error();
  ($main::verbosity > 2) && print "After ".tv_interval($t0)."s: Sent ".$nrPhrases." passphrases, gotEof=".$gotEof.", gotTimeout=".$gotTimeout.", lastError=".$lastError."\n";
  #$exp->soft_close();
  #print "DEBUG after ".tv_interval($t0)."s: hard_closing ...\n";
  $exp->hard_close();
  #print "DEBUG after ".tv_interval($t0)."s: hard_closed.\n";
  my $rpmExitcode = $exp->exitstatus();
  if (defined($rpmExitcode))  {
    if ($rpmExitcode) {
      print "rpm returned BAD exit code ".$rpmExitcode."\n";
      return 0;
    } else {
      ($main::verbosity > 2) && print "rpm returned with GOOD exit code ".$rpmExitcode."\n";
      return 1;
    }
  } else {
    print "rpm returned NO exit code\n";
    return 0;
  }

}


# new version stretch
# correctly sign rpm file, even unattended if correctly configured
# $1 = rpm file
sub signRpm( $ ) {
  my $rpmFile = shift;
  ($rpmFile eq "") && cleanupAndDie("FATAL: signRpm(\$) called with empty file spec.\n");

  # if we have to support pre-stretch environments we might do something like
  # if (! -e /usr/bin/rpmsign) { return signRpmJesse($rpmFile); }

  my $rpmCmd = "GNUPGHOME=".$main::gpgHome." ";
  $rpmCmd .= "rpmsign --addsign";
  if ($main::verbosity > 1) {
    $rpmCmd .= " -v";
  }
  if (length($main::gpgKeyId) > 0) {
    $rpmCmd .= " --key-id ".$main::gpgKeyId;
  }

  my ($rvRpm, $rvVerion, $rpmVersion) = split(/ /, `/usr/bin/rpmsign --version`);
  chomp $rpmVersion;
  ($main::verbosity > 2) && print "rpmsign version '".$rpmVersion."'.\n";

  if (length($rpmVersion) == 0 or (version_compare("4.14.2.1", $rpmVersion) > 0)) {
    ($main::verbosity > 2) && print "rpmsign version '".$rpmVersion."', using pre-buster parameters.\n";
  
    # this is needed after all, on builder:
    my $gpgAgentInfo = "";
    #if (defined($ENV{"GPG_AGENT_INFO"})) {
    #  $gpgAgentInfo = $ENV{"GPG_AGENT_INFO"};
    #}
    my $gpgUser = "";
    if (defined($ENV{"GPG_USER"})) {
      $gpgUser = $ENV{"GPG_USER"};
    }
    # gpg_sign_cmd
    $rpmCmd .= ' -D "%__gpg_sign_cmd %{__gpg} --no-verbose --no-armor'; # --passphrase-fd 3';
    #if (length($gpgAgentInfo) > 0) {
    #  $rpmCmd .= ' --use-agent --gpg-agent-info '.$gpgAgentInfo;
    #} else {
    #  $rpmCmd .= ' --no-use-agent';
    #}
    if (length($main::gpgHome) > 0) {
      $rpmCmd .= " --homedir ".$main::gpgHome;
    }
    if (length($main::gpgKeyId) > 0) {
      $rpmCmd .= " --default-key ".$main::gpgKeyId;
    }
    if (length($gpgUser) > 0) {
      $rpmCmd .= ' -u \"%{_gpg_name}\"';
    }
    $rpmCmd .= ' -sbo %{__signature_filename} %{__plaintext_filename}"';
    # gpg_name
    if (length($gpgUser) > 0) {
      $rpmCmd .= ' -D "%_gpg_name '.$gpgUser.'"';
    } else {
      $rpmCmd .= ' -D "%_gpg_name foobar"';
    }
  } else {
    ($main::verbosity > 2) && print "rpmsign version '".$rpmVersion."', using buster parameters.\n";
    $rpmCmd .= ' -D "%__gpg /usr/bin/gpg"';
  }

  $rpmCmd .= " ".$rpmFile;

  ($main::verbosity > 1) && print "signRpm(): Going to execute the following command:\n".$rpmCmd."\n\n";

  #executeOrDie($rpmCmd);
  system($rpmCmd);
  my $rpmExitcode = $? >> 8;
  if ($rpmExitcode == 0)  {
    ($main::verbosity > 2) && print "rpm returned with GOOD exit code ".$rpmExitcode."\n";
    return 1;
  } else {
    print "rpm returned BAD exit code ".$rpmExitcode."\n";
    return 0;
  }
}


# Other stuff


sub zipAndChmod ( $ ; $ ) {
  # $1 = file to zip (.gz and .bz2)
  # $2 = mode to chmod signed file and signature to; optional, defaults to 0755
  my $file = shift;
  ($file eq "") && cleanupAndDie("FATAL: zipAndChmod(\$[,\$]) called with empty file spec.\n");
  (-f $file) || cleanupAndDie("FATAL: zipAndChmod(".$file.",\$]): no such file.\n");
  my $mode = shift;
  if ($mode eq "") { $mode = 0755; }

  executeOrDie("cat ".$file." |gzip  >".$file.".gz");
  executeOrDie("cat ".$file." |bzip2 >".$file.".bz2");
  chmod $mode, $file, $file.".gz", $file.".bz2";
}


sub cleanupMounts() {
  if (scalar(@main::mountsToUnmount) > 0) {
    foreach my $mountPoint (@main::mountsToUnmount) {
      if ($main::verbosity > 0) { print "cleanup(): Unmounting ".$mountPoint."\n"; }
      system "/bin/umount ".$mountPoint;
    } 
  }
  @main::mountsToUnmount = ();
}

sub cleanup() {
  #if ($main::verbosity > 3) { print "cleanup(): pwd(): ".pwd()."\n"; }
  cleanupMounts();
  if (scalar(@main::filesToCleanup) > 0) {
    #if ($main::verbosity > 3) { 
    #  print "cleanup(): NOT rm'ing these files:\n".join("\n",@main::filesToCleanup)."\n"; 
    #} else {
      if ($main::verbosity > 1) { print "cleanup(): rm'ing these files:\n".join("\n",@main::filesToCleanup)."\n"; }
      #unlink @main::filesToCleanup or print "WARNING: cleanup(): Unable to unlink ".join(" ",@main::filesToCleanup)."\n";
      if ($main::verbosity > 2) {
        system "/bin/rm -fv ".join(" ", @main::filesToCleanup);
      } else {
        system "/bin/rm -f ".join(" ", @main::filesToCleanup);
      }
    #}
  }
  if (scalar(@main::dirsToCleanup) > 0) {
    #if ($main::verbosity > 3) { 
    #  print "cleanup(): NOT rm'ing these directories:\n".join("\n",@main::dirsToCleanup)."\n";
    #} else {
      if ($main::verbosity > 1) { print "cleanup(): rm'ing these directories:\n".join("\n",@main::dirsToCleanup)."\n"; }
      if ($main::verbosity > 2) {
        system "/bin/rm -rfv ".join(" ", @main::dirsToCleanup);
      } else {
        system "/bin/rm -rf ".join(" ", @main::dirsToCleanup);
      }
    #}
  }
}


sub cleanupAndDie( $ ) {
  # $1 = message
  my $message = shift;
  print "Cleanup before going to die for: ".$message."\n";
  cleanup();
  if ($message !~ /\n$/)
  { $message .= "\n"; }
  if ($message !~ /\n$/)
  { $message .= "\n"; }
  print "Dieing for: ".$message."\n";
  die $message;
}


# command helpers


sub executeOrDie ( $ ) {
  # $1 = command to execute
  # dies if command fails, i.e. if returncode != 0
  my $command = shift;
  ($command eq "") && cleanupAndDie("FATAL: executeOrDie(\$) called with empty command spec.\n");
  system($command);
  my $rc = $? >> 8;
  ($rc == 0) || cleanupAndDie("FATAL: Execution of command failed with returncode ".$rc.":\n".$command."\nCurrent directory is: ".pwd()."\n");
}

sub execute ( $ ) {
  # $1 = command to execute
  # if command fails (i.e. if returncode != 0), 0 (=false) is returned
  # otherwise 1 (true) is returned
  my $command = shift;
  ($command eq "") && cleanupAndDie("FATAL: execute(\$) called with empty command spec.\n");
  my $rc = system($command);
  ($rc == 0) && return 1; # true
  return 0; # false
}



# other helpers


sub myArchitecture() {
  my $architecture = "";
  #my $unameM = `uname -m`;
  #if ($unameM =~ /64/) {
  #  return "amd64";
  #}
  #if ($unameM =~ /i[3-6]86/) {
  #  return "i386";
  #}
  my $dpkgMainArch = `dpkg --print-architecture`;
  chomp $dpkgMainArch;
  if ($main::verbosity > 1) { print "myArchitecture(): Detected '".$dpkgMainArch."'\n"; }
  return $dpkgMainArch;
}


sub myUid() {
  my $uid = `id -u`;
  chomp $uid;
  ($uid eq "") && cleanupAndDie("FATAL: Cannot detect my uid using 'id -u'\n");
  return $uid;
}

sub dieIfNotRoot() {
  my $uid = myUid();
  #($uid eq "0") || cleanupAndDie("FATAL: Only root may do what you wanted.\n"));
  ($uid eq "0") || cleanupAndDie("FATAL: Only root may do what you wanted.\n");
}


# email functions


sub sendEMail( $ $ $ $ $ $ ) {
  # $1 = fromNameAdress (may be in format Name <adr@foo>), may be empty, use \s*,\s* to separate multiple addresses
  # $2 = toNamesAdresses (may be in format Name <adr@foo>), may be empty, use \s*,\s* to separate multiple addresses
  # $3 = ccNamesAdresses (may be in format Name <adr@foo>), may be empty, use \s*,\s* to separate multiple addresses
  # $4 = subject
  # $5 = message
  # $6 = namesOfFilesToAttach
  
  if ($main::smtpHost eq "-") {
    ($main::verbosity > 0) && print "sendEMail(): WARNING: Swallowing e-mail because STMPHost configured to '-'.\n\n"; 
    return;
  }
  my $headerFromNameAdr = shift;
  if ($headerFromNameAdr eq "") {
    $headerFromNameAdr = $ENV{"USER"}."@".$ENV{"HOSTNAME"};
  }
  my $listOfToNamesAdresses = shift;
  $listOfToNamesAdresses =~ s/(^\s+|\s+$)//g; 
  my $listOfCcNamesAdresses = shift;
  $listOfCcNamesAdresses =~ s/(^\s+|\s+$)//g; 
  my $subject = shift;
  my $message = shift;
  my $listOfFilesToAttach = shift;
  $listOfFilesToAttach =~ s/(^\s+|\s+$)//g; 
  my %recipients = ();
  my $headerTo = "";
  my $headerCc = "";
  # toNameAdr
  my @toNamesAdresses = split /\s*,\s*/, $listOfToNamesAdresses;
  foreach my $toNameAdr (@toNamesAdresses) {
    $toNameAdr =~ s/(^\s+|\s+$)//g; 
    my $toAdr=$toNameAdr;
    if ($toNameAdr =~ /.*(<[^>]+>)/) {
      $toAdr = $1;
    }
    $recipients{$toAdr} = $toAdr;
    if ($headerTo eq "") {
      $headerTo = $toNameAdr;
    } else {
      $headerTo .= ", ".$toNameAdr;
    }
  }
  # ccNameAdr
  my @ccNamesAdresses = split /\s*,\s*/, $listOfCcNamesAdresses;
  foreach my $ccNameAdr (@ccNamesAdresses) {
    $ccNameAdr =~ s/(^\s+|\s+$)//g; 
    my $ccAdr=$ccNameAdr;
    if ($ccNameAdr =~ /.*(<[^>]+>)/) {
      $ccAdr = $1;
    }
    $recipients{$ccAdr} = $ccAdr;
    if ($headerCc eq "") {
      $headerCc = $ccNameAdr;
    } else {
      $headerCc .= ", ".$ccNameAdr;
    }
  }
  # any recipients?
  if (scalar(keys(%recipients)) == 0)
  { print "sendEMail(): ERROR: No recipient specified, not sending anything.\n"; return; }
  # subject
  if ($subject eq "") 
  { print "sendEMail(): ERROR: No subject specified, not sending anything.\n"; return; }
  # 
  if ($message eq "") 
  { print "sendEMail(): ERROR: No message given, not sending anything.\n"; return; }

  my @filesToAttach = split /\s+/, $listOfFilesToAttach;
  my @emailParts = ();

  my $messagePart = Email::MIME->create(
    attributes => {
      content_type => "text/plain",
      charset => "UTF-8",
      encoding => "8bit",
    },
    body_str => $message
  );
  push @emailParts, $messagePart;

  foreach my $fileToAttach (@filesToAttach) {
    $fileToAttach =~ s/(^\s+|\s+$)//g; 
    if (not -f $fileToAttach) {
      print "ERROR: Cannot read file to attach: ".$fileToAttach."\n";
      next;
    }
    # gz most stuff
    if ($fileToAttach =~ /\.(log|txt)$/) {
      executeOrDie("gzip ".$fileToAttach);
      $fileToAttach .= ".gz";
    }
    #print "DEBUG: Attaching file: ".$fileToAttach."\n";
    my @fileNameParts = split /\//, $fileToAttach;
    my $baseName = pop @fileNameParts;
    #$baseName =~ s/-$$//g;
    if ($fileToAttach =~ /\.(log|txt)$/) {
      #print "DEBUG: Treating as text\n";
      my $attachmentPart = Email::MIME->create(
        attributes => { 
          content_type => "text/plain",
          charset => "UTF-8",
          encoding => "8bit",
          filename => $baseName,
          name => $baseName
        },
        body_str => readCompleteFile( $fileToAttach )
      );
      push @emailParts, $attachmentPart;
      next;
    }
    if ($fileToAttach =~ /\.gz$/) {
      #print "DEBUG: Treating as gzip\n";
      my $attachmentPart = Email::MIME->create(
        attributes => { 
          content_type => "application/x-gzip",
          encoding => "base64",
          filename => $baseName,
          name => $baseName
        },
        body => io($fileToAttach)->all()
      );
      push @emailParts, $attachmentPart;
      next;
    }
    if ($fileToAttach =~ /\.bz2$/) {
      #print "DEBUG: Treating as gzip\n";
      my $attachmentPart = Email::MIME->create(
        attributes => { 
          content_type => "application/x-bzip",
          encoding => "base64",
          filename => $baseName,
          name => $baseName
        },
        body => io($fileToAttach)->all()
      );
      push @emailParts, $attachmentPart;
      next;
    }
    # else how Email::MIME is good enough
    #print "DEBUG: Treating as other\n";
    my $attachmentPart = Email::MIME->create(
      attributes => { 
        content_type => "application/octet-stream",
        encoding => "base64",
        filename => $baseName,
        name => $baseName,
      },
      body => io($fileToAttach)->all
    );
    push @emailParts, $attachmentPart;
  }

  my $email = Email::MIME->create(
    header => [
      From => $headerFromNameAdr,
      To => $headerTo,
      Cc => $headerCc,
      Subject => $subject ],
    parts => [ @emailParts ]
  );
 
  my $doUseTls = 0;
  if ($main::smtpPort == 465) {
    $doUseTls = 1;
  }
  my $doDebugSmtp = 0;
  if ($main::verbosity > 2) {
    $doDebugSmtp = 1;
  }
  my $smtp = Net::SMTP->new(
      Host => $main::smtpHost,
      Port => $main::smtpPort,
      SSL => $doUseTls,
      Debug => $doDebugSmtp,
      Hello => $main::smtpEhlo,
      Timeout => 30
    );
  if (!defined($smtp)) {
    print "ERROR: SMTP object undefined, not sending any e-mails. Failure reason: ".$@."\n";
    return;
  }
  if (length($main::smtpUser) > 0) {
    $smtp->auth( $main::smtpUser, $main::smtpPassword );
  }
  $smtp->mail($main::buildReportsFrom);
  $smtp->recipient(sort(keys(%recipients)));
  $smtp->data($email->as_string());
  $smtp->quit();

}


# script creation functions

sub proposeCpScript( $ $ ) {
  # $1 = targetDirectory
  # $2 = space sep. list of files to cp
  my $targetDirectory = shift;
  ($targetDirectory eq "") && cleanupAndDie("FATAL: proposeCpScript called without targetDirectory as \$1.\n");
  my $fileList = shift;
  ($fileList eq "") && cleanupAndDie("FATAL: proposeCpScript called without fileList as \$2.\n");
  my @filesToCp = split / /, $fileList;
  # input params ok
  my $feedScript;
  $feedScript .= "#!/bin/bash\n";
  $feedScript .= "/bin/cp -v \\\n";
  foreach my $fileToCp (@filesToCp) {
    $feedScript .= "  ".$fileToCp." \\\n";
  }
  $feedScript .= "  ".$targetDirectory."/ \\\n";
  $feedScript .= "  || exit 1\n";
  #$feedScript .= "exit 0\n";
  $feedScript .= "\n";
  return $feedScript;
}

sub proposeFeedScript( $ $ $ ) {
  # $1 = feedHostDir
  # $2 = uploadName
  # $3 = space sep. list of files to upload
  my $feedHostDir = shift;
  ($feedHostDir eq "") && cleanupAndDie("FATAL: proposeFeedScript called without feedHostDir as \$1.\n");
  my $uploadName = shift;
  ($uploadName eq "") && cleanupAndDie("FATAL: proposeFeedScript called without uploadName as \$2.\n");
  my $fileList = shift;
  ($uploadName eq "") && cleanupAndDie("FATAL: proposeFeedScript called without fileList as \$3.\n");
  my @filesToUpload = split / /, $fileList;
  # input params ok
  my ($feedHost, $feedDir) = split /:/, $feedHostDir;
  my $uploadDir = $feedDir."/uploading-".$uploadName."_".$$;
  my $feedScript;
  $feedScript .= "#!/bin/bash\n";
  $feedScript .= "/usr/bin/ssh ".$feedHost." /bin/mkdir -pv ".$uploadDir." \\\n";
  $feedScript .= "  || /usr/bin/ssh ".$feedHost." /bin/mkdir -pv ".$uploadDir." \\\n";
  $feedScript .= "  || exit 2\n";
  $feedScript .= "/usr/bin/scp \\\n";
  foreach my $fileToUpload (@filesToUpload) {
    $feedScript .= "  ".$fileToUpload." \\\n";
  }
  $feedScript .= "  ".$feedHost.":".$uploadDir."/ \\\n";
  $feedScript .= "  || /usr/bin/scp \\\n";
  foreach my $fileToUpload (@filesToUpload) {
    $feedScript .= "  ".$fileToUpload." \\\n";
  }
  $feedScript .= "  ".$feedHost.":".$uploadDir."/ \\\n";
  $feedScript .= "  || exit 3\n";
  $feedScript .= "/usr/bin/ssh ".$feedHost." /bin/mv -v ".$uploadDir."/* ".$feedDir."/ \\\n";
  $feedScript .= "  || /usr/bin/ssh ".$feedHost." /bin/mv -v ".$uploadDir."/* ".$feedDir."/ \\\n";
  $feedScript .= "  || exit 4\n";
  $feedScript .= "/usr/bin/ssh ".$feedHost." /bin/rmdir -v ".$uploadDir." \\\n";
  $feedScript .= "  || /usr/bin/ssh ".$feedHost." /bin/rmdir -v ".$uploadDir." \\\n";
  #$feedScript .= "  || exit 5\n";
  $feedScript .= "  || echo 'WARNING: Failed to rm empty upload dir on repository host'\n";
  #$feedScript .= "exit 0\n";
  $feedScript .= "\n";
  return $feedScript;
}

sub proposeMvScript( $ $ ) {
  # $1 = targetDirectory
  # $2 = space sep. list of files to mv
  my $targetDirectory = shift;
  ($targetDirectory eq "") && cleanupAndDie("FATAL: proposeMvScript called without targetDirectory as \$1.\n");
  my $fileList = shift;
  ($fileList eq "") && cleanupAndDie("FATAL: proposeMvScript called without fileList as \$2.\n");
  my @filesToMv = split / /, $fileList;
  # input params ok
  my $feedScript;
  $feedScript .= "#!/bin/bash\n";
  $feedScript .= "/bin/mv -v \\\n";
  foreach my $fileToMv (@filesToMv) {
    $feedScript .= "  ".$fileToMv." \\\n";
  }
  $feedScript .= "  ".$targetDirectory."/ \\\n";
  $feedScript .= "  || exit 1\n";
  #$feedScript .= "exit 0\n";
  $feedScript .= "\n";
  return $feedScript;
}



# EOF
1;

